import functools


# Subscriptions map
# Topic -> fn
subscriptions = dict()


def subscribe(topic: str):
    """Relates the topic handler to the specified topic.

    This will automatically subscribe to the topic on reconnects.

    @mqtt.subscribe('/my/topic')
    def handle_topic(client, userdata, message):
        pass
    """
    def decorator(fn):
        subscriptions[topic] = fn

        @functools.wraps(fn)
        def add_subscription(*args, **kwargs):
            return fn(*args, **kwargs)
        return add_subscription
    return decorator



import functools

from paho.mqtt.client import Client


on_connects = list()


def on_connect():
    """Excutes the function while injecting a connected MQTT client.
        This function will always be called when the client raises the
        on_connect function.

        Usage:

        @mqtt.on_connect()
        def init_publish(client, *args, **kwargs):
            client.publish('/here', 'Me!')
    """
    def decorator(fn):
        on_connects.append(fn)

        @functools.wraps(fn)
        def add_on_connect(client: Client, *args, **kwargs):
            return fn(client=client, *args, **kwargs)
        return add_on_connect
    return decorator



import logging
import sys


LOGGING_CONFIG_DEFAULTS = dict(
    version=1,
    disable_existing_loggers=False,

    loggers={
        'pahotoolkit': {
            'level': 'DEBUG',
            'handlers': ['console']
        },
        'pahotoolkit_error': {
            'level': 'INFO',
            'handlers': ['error_console'],
            'propagate': True,
            'qualname': 'error'
        },
    },
    handlers={
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'generic',
            'stream': sys.stdout
        },
        'error_console': {
            'class': 'logging.StreamHandler',
            'formatter': 'generic',
            'stream': sys.stderr
        },
    },
    formatters={
        'generic': {
            'format': '[%(process)d - %(thread)d - %(module)s - %(funcName)s] [%(levelname)s] %(message)s',  # noqa
            'class': 'logging.Formatter'
        },
    }
)

logger = logging.getLogger('pahotoolkit')

error_logger = logging.getLogger('pahotoolkit_error')



"""MQTT simplified client module."""
import logging.config

from paho.mqtt.client import Client

from pahotoolkit import exc
from pahotoolkit.log import (
    LOGGING_CONFIG_DEFAULTS,
    logger,
    error_logger,
)
from pahotoolkit.subscriptions import subscriptions
from pahotoolkit.on_connect import on_connects


# singleton client
client = Client()


def start_async(host: str, port: int = 1883,
                username: str = None, password: str = None,
                initialize: bool = True,
                log_config: dict = LOGGING_CONFIG_DEFAULTS,
                mqtt_client: Client = client) -> Client:
    """Start the MQTT Client in async mode."""
    if not host:
        raise exc.InvalidMqttHost('An MQTT host is required')

    if username and password:
        mqtt_client.username_pw_set(username=username,
                                    password=password)

    if initialize:
        init(mqtt_client)

    mqtt_client.connect_async(host=host,
                              port=port)
    mqtt_client.loop_start()

    if log_config:
        logging.config.dictConfig(log_config)

    return mqtt_client


def stop_async(mqtt_client: Client = client):
    """Stops the client's loop."""
    mqtt_client.loop_stop()


def init(mqtt_client: Client = client):
    """Initializes the given Paho's MQTT Client.

    If the given MQTT client is not given then it uses the
    singleton MQTT client.
    """
    mqtt_client.on_connect = _on_connect
    mqtt_client.on_message = _on_message


def _on_connect(client: Client, userdata, flags, result_code):
    """Handle the on-connect event from paho's client."""
    if result_code == 0:
        logger.debug(exc.OK)
        hook(client)
    elif result_code == 1:
        error_logger.error(exc.CONN_INCORRECT_PROTOCOL_VER)
    elif result_code == 2:
        error_logger.error(exc.CONN_INVALID_CLIENT_ID)
    elif result_code == 3:
        error_logger.error(exc.SERVER_UNAVAILABLE)
    elif result_code == 4:
        error_logger.error(exc.INVALID_CREDENTIALS)
    elif result_code == 5:
        error_logger.error(exc.UNAUTHORIZED)
    else:
        error_logger.warning(exc.UNKNOWN)


def hook(mqtt_client: Client):
    """Hook the toolkit to Paho's MQTT Client."""
    mqtt_client.subscribe('$SYS/#')

    logger.debug('Registering annotated subscriptions')
    # register the message callback for the topics
    for topic, fn in subscriptions.items():
        mqtt_client.subscribe(topic)
        mqtt_client.message_callback_add(topic,
                                         callback=fn)

    logger.debug('Registering annotated on_connect hooks')
    # register the on_connect functions
    for fn in on_connects:
        try:
            fn(client=mqtt_client)
        except Exception as e:
            err_message = 'on_connect function failed: {e}'.format(e=str(e))
            error_logger.error(err_message)


def _on_message(client: Client, userdata, msg):
    """Unhandled message handler."""
    log_message = 'Unhandled message, {topic}: {data}'.format(topic=msg.topic,
                                                              data=msg.payload)
    error_logger.warning(log_message)



"""Exceptions."""

# on_connection code descriptions #
OK = 'MQTT client connected'

CONN_INCORRECT_PROTOCOL_VER = 'Connection refused - incorrect protocol version'

CONN_INVALID_CLIENT_ID = 'Connection refused - Invalid client identifier'

SERVER_UNAVAILABLE = 'Connection refused - Server Unavailable'

INVALID_CREDENTIALS = 'Connection refused - Bad username or password'

UNAUTHORIZED = 'Connection refused - Not Authorized'

UNKNOWN = 'Unknown connection code'


class PahoToolKitException(Exception):
    """Base Paho's Toolkit exception."""


class InvalidMqttHost(PahoToolKitException):
    """The given MQTT host is invalid."""



import functools
import json

from paho.mqtt.client import MQTTMessage

from pahotoolkit.log import error_logger


def json_message(none_on_failure: bool = True, json_loader=json):
    """Json Message or nothing, maybe.

    This decorator will filter all the non-JSON
    incoming messages. If the :param none_on_failure:
    is set, it will return None, avoiding the call of
    the wrapped function.

    If set to False the function will always get called,
    but the payload argument will be always an empty
    dictionary for non-JSON messages.

    The json_loader argument allows to specify different a
    json module to use for calling the `loads` method.
    This is useful when requiring an alternative to the
    standard `json` module.

    :param none_on_failure:
    :param json_loader:
    :params args:
    :params kwargs:

    :kwarg client, userdata, message: MQTT Paho original
    arguments

    Usage:

    @json_message()
    def handle_message(payload, *args, **kwargs):
        pass
    """
    def decorator(fn):

        @functools.wraps(fn)
        def execute(client, userdata, message: MQTTMessage, *args, **kwargs):
            payload = {}
            try:
                payload = json_loader.loads(message.payload)
            except Exception as e:
                if none_on_failure:
                    err_message = 'Message was not JSON, {e}'.format(e=str(e))
                    error_logger.error(err_message)
                    return None

            return fn(client=client, userdata=userdata, message=message,
                      payload=payload, *args, **kwargs)
        return execute
    return decorator



"""Paho's Toolkit index."""

from pahotoolkit.client import (
    client,
    init,
    start_async,
    stop_async,
)
from pahotoolkit.exc import (
    PahoToolKitException,
    InvalidMqttHost,
)
from pahotoolkit.log import (
    logger,
    error_logger
)
from pahotoolkit.on_connect import on_connect
from pahotoolkit.subscriptions import subscribe
from pahotoolkit.utils import json_message


__all__ = [
    'client', 'init', 'start_async', 'stop_async',
    'logger', 'error_logger',
    'PahoToolKitException', 'InvalidMqttHost',
    'subscribe', 'on_connect', 'json_message',
]



import codecs
import os

from setuptools import setup


__version__ = os.getenv('TAG', '0.1.4.2')


URL = 'https://bitbucket.org/kiotsystem/paho-toolkit'

download_url = '{url}/get/{version}.tar.gz'.format(url=URL,
                                                   version=__version__)


def open_local(paths, mode='r', encoding='utf8'):
    path = os.path.join(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            *paths
        )
    )

    return codecs.open(path, mode, encoding)


with open_local(['README.md']) as readme:
    long_description = readme.read()


with open_local(['requirements.txt']) as req:
    install_requires = req.read().split("\n")


keywords = ['paho', 'eclipse', 'mqtt', 'toolkit', 'framework']


# http://peterdowns.com/posts/first-time-with-pypi.html
# https://bitbucket.org/kiotsystem/paho-toolkit/get/0.1b2.tar.gz
setup(
    name='pahotoolkit',
    packages=['pahotoolkit'],
    version=__version__,
    description='Paho\'s MQTT dev Toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Arnulfo Solis',
    author_email='arnulfojr94@gmail.com',
    url=URL,
    download_url=download_url,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=install_requires,
    keywords=keywords,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
